import sys
from abc import abstractmethod

import requests
import numpy as np
from foolbox.models import Model
import bson

from .retry_helper import retryable
from .logger import logger
from .common import check_image

if sys.version_info > (3, 3):
    import urllib.parse as parse
else:
    import urlparse as parse


class HTTPClient(object):
    """Base class for HTTPModel and HTTPAttack."""

    def _encode_array_data(self, array):
        """
        Converts a numpy array to bytes.
        """
        return array.tobytes()

    def _decode_array_data(self, data, dtype, shape):
        """
        Converts bytes to a numpy array.
        """
        return np.frombuffer(data, dtype=dtype).reshape(shape)

    def _encode_arrays(self, data):
        """
        Can be used by subclasses to encode numpy arrays. To use it,
        subclasses must implement _encode_array_data.
        """
        encoded = {}
        for key in list(data.keys()):
            if isinstance(data[key], np.ndarray):
                array = data[key]
                encoded[key] = {
                    'type': 'array',
                    'shape': array.shape,
                    'dtype': array.dtype.str,
                    'data': self._encode_array_data(array),
                }
            else:
                encoded[key] = data[key]
        return encoded

    def _decode_arrays(self, encoded):
        """
        Can be used by subclasses to decode numpy arrays. To use it,
        subclasses must implement _decode_array_data.
        """
        decoded = {}
        for key in list(encoded.keys()):
            if hasattr(encoded[key], 'get') \
                    and encoded[key].get('type') == 'array':
                shape = encoded[key]['shape']
                dtype = encoded[key]['dtype']
                data = encoded[key]['data']
                array = self._decode_array_data(data, dtype, shape)
                decoded[key] = array
            else:
                decoded[key] = encoded[key]
        return decoded

    @retryable
    def _post(self, path, data):
        """
        Encodes the data dictionary, sends it to the http server using
        an http post request to the url specified by path, decodes
        the result and returns it as a dictionary.
        """
        url = self._url(path=path)
        headers = {'content-type': 'application/bson'}
        data = self._encode_arrays(data)
        data = bson.dumps(data)
        r = self.requests.post(url, headers=headers, data=data)
        r.raise_for_status()
        assert r.ok
        result = r.content
        result = bson.loads(result)
        result = self._decode_arrays(result)
        return result

    @retryable
    def _get(self, path):
        """
        Performs a get request to the url specified by path and
        returns the result as text. Subclasses can override this
        if necessary.
        """
        url = self._url(path=path)
        r = self.requests.get(url)
        r.raise_for_status()
        assert r.ok
        return r.text

    @abstractmethod
    def _url(self, path=''):
        raise NotImplementedError


class TinyImageNetBSONModel(Model, HTTPClient):
    """Base class for models that connect to an http server and
    dispatch all requets to that server.

    Parameters
    ----------
    url : str
        The http or https URL of the server.

    """

    def __init__(self, url):
        self.requests = requests

        self._base_url = url

        super(TinyImageNetBSONModel, self).__init__(
            bounds=(0, 255), channel_axis=3)

    def _url(self, path=''):
        return parse.urljoin(self._base_url, path)

    @property
    def base_url(self):
        return self._base_url

    def server_version(self):
        return self._get('/server_version')

    def __call__(self, image):
        return self.predict(image)

    def predict(self, image):
        image = check_image(image)
        data = {'image': image}
        result = self._post('/predict', data)

        prediction = result['prediction']
        assert isinstance(prediction, int), "prediction should return an int value, but got: %s" % type(prediction)
        assert (0 <= prediction < 200), "prediction should be a value between 0 and 200, but got: %s" % prediction
        return prediction

    def batch_predictions(self, images):
        assert images.shape[0] == 1
        image = images[0]
        predictions = self.predictions(image)
        predictions = predictions[np.newaxis]
        return predictions

    def predictions(self, image):
        class_ = self.predict(image)
        predictions = np.zeros((200,), dtype=np.float32)
        predictions[class_] = 1
        return predictions

    def num_classes(self):
        return 200



import numpy as np
from .logger import logger
import json
import os


def check_image(image):
    # image should a 64 x 64 x 3 RGB image
    assert isinstance(
        image, np.ndarray), "image should be an numpy array"
    assert image.shape == (64, 64, 3), "image should be of size 64x64x3"
    if image.dtype == np.float32:
        # we accept float32, but only if the values
        # are between 0 and 255 and we convert them
        # to integers
        if image.min() < 0:
            logger.warning('clipped value smaller than 0 to 0')
        if image.max() > 255:
            logger.warning('clipped value greater than 255 to 255')
        image = np.clip(image, 0, 255)
        image = image.astype(np.uint8)
    assert image.dtype == np.uint8
    return image

def check_track(directory, track):
    crowdai_json = os.path.join(directory, "crowdai.json")
    with open(crowdai_json) as file:
        data = json.load(file)
    id = data['challenge_id']

    assert track == id, "your running test script for {0}, but the crowdai.json says: {1}".format(track, id)



import os

import numpy as np
import yaml

from .client import TinyImageNetBSONModel
from .logger import logger
from .notifier import CrowdAiNotifier
from .common import check_image

from adversarial_vision_challenge.retry_helper import RetriesExceededError


def _load_img(path):
    """
        Reads image image from the given path and returns an numpy array.
    """
    path = os.path.join(os.path.dirname(__file__), path)
    image = np.load(path)
    assert image.dtype == np.uint8
    assert image.shape == (64, 64, 3)
    return image


def _read_image(file_name):
    """
        Returns a tuple of image as numpy array and label as int,
        given the csv row.
    """
    input_folder = os.getenv('INPUT_IMG_PATH')
    img_path = os.path.join(input_folder, file_name)
    image = _load_img(img_path)
    assert image.dtype == np.uint8
    image = image.astype(np.float32)
    assert image.dtype == np.float32
    return image


def read_images():
    """
        Returns a list containing tuples of images as numpy arrays
        and the correspoding label.
        In case of an untargeted attack the label is the ground truth label.
        In case of a targeted attack the label is the target label.
    """
    filepath = os.getenv('INPUT_YML_PATH')
    with open(filepath, 'r') as ymlfile:
        data = yaml.load(ymlfile)

    return [(key, _read_image(key), data[key]) for key in data.keys()]


def store_adversarial(file_name, adversarial):
    """
        Given the filename, stores the adversarial as .npy file.
    """
    if adversarial is not None:
        adversarial = check_image(adversarial)

    output_folder = os.getenv('OUTPUT_ADVERSARIAL_PATH')
    path = os.path.join(output_folder, file_name)
    path_without_extension = os.path.splitext(path)[0]
    np.save(path_without_extension, adversarial)
    CrowdAiNotifier.store_adversarial(file_name)


def attack_complete():
    """
        Send a notificaton to the crowd-ai backend that the attack has
        successfully completed.
    """
    CrowdAiNotifier.attack_complete()


def _wait_for_server_start(model, retried=0):
    version = ''
    try:
        version = model.server_version()
    except RetriesExceededError:
        if 'NIPS 2018' not in version:
            if retried < 3:
                _wait_for_server_start(model, retried + 1)
            else:
                logger.error("Can't reach model server: %s.", model.base_url)


def load_model():
    """
        Returns an BSONModel reading the server URI and post from
        environment variables.
    """
    model_port = os.getenv('MODEL_PORT', 8989)
    model_server = os.getenv('MODEL_SERVER', 'localhost')
    model_url = 'http://{0}:{1}'.format(model_server, model_port)
    model = TinyImageNetBSONModel(model_url)
    _wait_for_server_start(model)
    return model


def get_test_data():
    """
        Returns a list with entries (image, label) over test samples.
    """
    basepath = os.path.join(os.path.dirname(__file__), 'test_images/')
    label_file = os.path.join(basepath, 'labels.yml')
    with open(label_file, 'r') as ymlfile:
        files2labels = yaml.load(ymlfile)

    return [(_load_img(os.path.join('test_images', filename)), label)
            for filename, label in sorted(files2labels.items())]



import crowdai_api

from enum import Enum


class ModelNotifications:
    TYPE = "AVC.MODEL"
    TOO_MANY_REQUESTS = "AVC.MODEL.TOO_MANY_REQUESTS_ERROR"


class AttackNotifications:
    TYPE = "AVC.ATTACK"
    COMPLETE = "AVC.ATTACK.COMPLETE"
    RETRIES_EXCEEDED = "AVC.ATTACK.RETRIES_EXCEEDED_ERROR"
    STORE_ADVERSARIAL = "AVC.ATTACK.STORE_ADVERSARIAL"

class CrowdAiNotifier():

    @staticmethod
    def _send_notification(event_type, message, payload={}):
        crowdai_events = crowdai_api.events.CrowdAIEvents()
        default_payload = {"challenge_id": "NIPS18_AVC"}
        default_payload.update(payload)
        crowdai_events.register_event(event_type, message, payload)

    # ~~~~~~~~~~~~~~~~ ATTACK NOTIFICATIONS ~~~~~~~~~~~~~~~~
    @staticmethod
    def store_adversarial(filename):
        CrowdAiNotifier._send_notification(
            event_type=AttackNotifications.STORE_ADVERSARIAL,
            message="",
            payload={
                "type": AttackNotifications.TYPE,
                "filename" : filename
            }
        )

    @staticmethod
    def attack_complete():
        CrowdAiNotifier._send_notification(
            event_type=AttackNotifications.COMPLETE,
            message="Attack successfully completed.",
            payload={
                "type": AttackNotifications.TYPE
            }
        )

    @staticmethod
    def retries_exceeded():
        CrowdAiNotifier._send_notification(
            event_type=AttackNotifications.RETRIES_EXCEEDED,
            message="No proper response from model after retrying multiple times.",
            payload={
                "type": AttackNotifications.TYPE
            }
        )

    # ~~~~~~~~~~~~~~~~ MODEL NOTIFICATIONS ~~~~~~~~~~~~~~~~

    @staticmethod
    def too_many_requests():
        CrowdAiNotifier._send_notification(
            event_type=ModelNotifications.TOO_MANY_REQUESTS,
            message="The attack has exceeded the max number of allowed predictions requests.",
            payload={
                "type": ModelNotifications.TYPE
            }
        )



import logging
import os

filename = os.getenv('LOG_FILE')
logger = None

if filename is not None:
    FORMAT = '%(asctime)-15s %(message)s'
    logging.basicConfig(filename=filename, format=FORMAT)
    logger = logging.getLogger('adversarial_vision_challenge')
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(logging.INFO)
else:
    logger = logging.getLogger()
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(logging.INFO)




import requests
from functools import wraps
import time
from .logger import logger
import traceback
import sys
from .notifier import CrowdAiNotifier

class RetriesExceededError(Exception):
    pass


def retryable(func, retries=3):
    @wraps(func)
    def retry(*args, **kwargs):
        for retried in range(retries + 1):
            if retried > 0:
                logger.info('Retrying for the %s. time', retried)
                time.sleep(3 * retried)
            try:
                return func(*args, **kwargs)
            except requests.exceptions.RequestException:
                traceback.print_exc(file=sys.stdout)
        logger.error('Retried request for %s times. Giving up.', retried)
        CrowdAiNotifier.retries_exceeded()
        raise RetriesExceededError(
            "Failed already {0} times. No further retrying.".format(retries))
    return retry



from __future__ import absolute_import
from __future__ import print_function

import inspect
import os
from functools import wraps
from io import BytesIO
import timeit

import bson
import numpy as np
from flask import Flask, Response, request
from PIL import Image

# from werkzeug.exceptions import BadRequest
from werkzeug.exceptions import TooManyRequests

from . import __version__
from .logger import logger
from .notifier import CrowdAiNotifier


# the number of max requests to predict for this model run
# based on the # of images to predict
number_of_max_predictions = int(os.environ.get('NUM_OF_MAX_PREDICTIONS', 100)) * 1000


def model_server(model):
    """Starts an HTTP server that provides access to a Foolbox model.

    Parameters
    ----------
    model : `foolbox.model.Model` instance
        The model that should be run.
    port : int
        The TCP port used by the HTTP server. Defaults to the MODEL_PORT environment
        variable or 8989 if not set.

    """


    port = int(os.environ.get('MODEL_PORT', 8989))

    app = Flask(__name__)

    # disable verbose flask loggig
    import logging
    log = logging.getLogger('werkzeug')
    log.setLevel(logging.ERROR)

    channel_axis = model.channel_axis()
    assert channel_axis in [1, 3], "model channel axis should be either 1 or 3"

    try:
        bounds = model.bounds()
    except AttributeError:
        bounds = (0, 255)
        logger.info('model has no bounds method, assuming (0, 255)')

    assert bounds == (0, 255), (
        'bounds must be (0, 255), update your model or use the preprocessing '
        'argument of foolbox model wrappers')

    def _predict(image):
        assert isinstance(image, np.ndarray), "input image should be an numpy array"
        assert image.shape == (64, 64, 3), "input image should be of size 64x64x3"
        assert image.dtype == np.uint8, "image should be of type np.uint8, but got: %s" % image.dtype

        # models (should) expect float32 arrays
        image = image.astype(np.float32)

        if channel_axis == 1:
            image = np.transpose(image, [2, 0, 1])

        prediction = model.predictions(image)

        if isinstance(prediction, np.ndarray) and prediction.size > 1:
            assert prediction.size == 200, "prediction.size should be 200, but got: %s" % prediction.size
            prediction = np.argmax(prediction)
        prediction = int(prediction)
        assert 0 <= prediction < 200, "prediction should be a value between 0 and 200, but got: %s" % prediction
        return prediction

    _predict = _wrap(_predict, ['prediction'])

    @app.route("/")
    def main():  # pragma: no cover
        return Response(
            'NIPS 2018 Adversarial Vision Challenge Model Server\n',
            mimetype='text/plain')

    @app.route("/server_version", methods=['GET'])
    def server_version():
        v = __version__
        return Response(str(v), mimetype='text/plain')

    @app.route("/predict", methods=['POST'])
    def predict():
        _check_rate_limitation()
        start = timeit.default_timer()
        prediction = _predict(request)
        end = timeit.default_timer()
        logger.debug('prediction took: %s s', (end - start))
        return prediction

    @app.route("/shutdown", methods=['GET'])
    def shutdown():
        _shutdown_server()
        return 'Shutting down ...'

    logger.info('starting server on port {}'.format(port))
    app.run(host='0.0.0.0', port=port)


def _check_rate_limitation():
    global number_of_max_predictions
    logger.debug('Number of remaining max requests: %s',
                 number_of_max_predictions)
    number_of_max_predictions -= 1
    if (number_of_max_predictions < 0):
        logger.error('Maximal number of prediction requests exceeded: %s',
                     number_of_max_predictions)
        CrowdAiNotifier.too_many_requests()
        raise TooManyRequests(
            'Maximal number of prediction requests exceeded: {0}'.format(
                number_of_max_predictions))


def _shutdown_server():
    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:  # pragma: no cover
        raise RuntimeError('Not running with the Werkzeug Server')
    func()


def _wrap(function, output_names):
    """A decorator that converts data between flask and python / numpy"""

    try:
        # Python 3
        sig = inspect.signature(function)
        params = sig.parameters
    except AttributeError:  # pragma: no cover
        # Python 2.7
        argspec = inspect.getargspec(function)
        params = dict(zip(argspec.args, [None] * len(argspec.args)))

    @wraps(function)
    def wrapper(request):
        verbose = request.args.get('verbose', False)

        if verbose:  # pragma: no cover
            print('headers', request.headers)
            print('args', list(request.args.keys()))
            print('form keys', list(request.form.keys()))
            print('files', list(request.files.keys()))
            print('is_json', request.is_json)
            print('data length', len(request.data))

        content_type = request.headers.get('content-type', '').lower()

        if content_type == 'application/bson':
            bson_args = bson.loads(request.data)
            bson_args = _decode_arrays(bson_args)

        else:  # pragma: no cover
            bson_args = {}

        args = {}

        def add_argument(name, value):
            if name in args:  # pragma: no cover
                print('ignoring {}, argument already exists'.format(name))
                return
            if name not in params:  # pragma: no cover
                print('ignoring {}, not accepted by function'.format(name))
                return
            args[name] = value

        for name, value in bson_args.items():
            add_argument(name, value)

        for name, value in request.args.items():  # pragma: no cover
            add_argument(name, value)

        for name, value in request.form.items():  # pragma: no cover
            add_argument(name, value)

        for name, value in request.files.items():  # pragma: no cover
            if name not in params:
                continue
            data = value.read()
            param = params[name]
            if param is not None and param.annotation == Image.Image:
                data = Image.open(BytesIO(data))
            add_argument(name, data)

        result = function(**args)
        if len(output_names) == 1:
            result = {output_names[0]: result}
        else:
            assert len(result) == len(output_names)
            result = dict(zip(output_names, result))
        result = _encode_arrays(result)
        result = bson.dumps(result)
        return Response(result, mimetype='application/bson')

    return wrapper


def _encode_arrays(d):
    for key in list(d.keys()):
        if isinstance(d[key], np.ndarray):
            array = d[key]
            d[key] = {
                'type': 'array',
                'shape': array.shape,
                'dtype': array.dtype.str,
                'data': array.tobytes(),
            }
    return d


def _decode_arrays(d):
    for key in list(d.keys()):
        if hasattr(d[key], 'get') \
                and d[key].get('type') == 'array':
            shape = d[key]['shape']
            dtype = d[key]['dtype']
            data = d[key]['data']
            array = np.frombuffer(data, dtype=dtype).reshape(shape)
            d[key] = array
    return d



from setuptools import setup
from setuptools import find_packages
from os.path import join, dirname
# We need io.open() (Python 3's default open) to specify file encodings
import io
import sys

with open(join(dirname(__file__), 'adversarial_vision_challenge/VERSION')) as f:
    version = f.read().strip()

try:
    # obtain long description from README and CHANGES
    # Specify encoding to get a unicode type in Python 2 and a str in Python 3
    with io.open(
            join(dirname(__file__), 'README.rst'),
            'r',
            encoding='utf-8') as f:  # type: ignore
        README = f.read()
except IOError:
    README = ''

install_requires = [
    'bson',
    'flask',
    'foolbox',
    'numpy',
    'pillow',
    'requests',
    'setuptools',
    'pyyaml',
    'crowdai_api',
    'GitPython',
    'packaging',
    'future',
    "crowdai-repo2docker ; python_version>='3.4'",
    'tqdm'
]

tests_require = [
    'pytest',
    'pytest-cov',
]

setup(
    name="adversarial_vision_challenge",
    version=version,
    description="Tools for the NIPS Adversarial Vision Challenge. Includes an HTTP server and client that provides access to Foolbox models and attacks.",  # noqa: E501
    long_description=README,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    keywords="",
    author="Jonas Rauber & Wieland Brendel & Behar Veliqi",
    author_email="behar.veliqi@uni-tuebingen.de",
    url="https://github.com/bethgelab/adversarial-vision-challenge",
    license="MIT",
    packages=find_packages(),
    scripts=[
        'bin/avc-test-setup',
        'bin/avc-test-model',
        'bin/avc-test-attack',
        'bin/avc-test-untargeted-attack',
        'bin/avc-test-targeted-attack',
        'bin/avc-submit'
    ],
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    extras_require={
        'testing': tests_require,
    },
)



from os.path import join as _join
from os.path import dirname as _dirname

with open(_join(_dirname(__file__), 'VERSION')) as _f:
    __version__ = _f.read().strip()

from .server import model_server  # noqa: F401
from .client import TinyImageNetBSONModel  # noqa: F401
from .utils import load_model  # noqa: F401
from .utils import read_images  # noqa: F401
from .utils import store_adversarial  # noqa: F401
from .utils import get_test_data  # noqa: F401
from .utils import attack_complete # noqa: F401
from .notifier import ModelNotifications, AttackNotifications # noqa: F401
